varying highp vec2 coordinate;
uniform sampler2D framePrev;
uniform sampler2D frameNext;
uniform sampler2D motionTexture;
uniform highp float deltaX;
uniform highp float deltaY;


const highp float textureUniformEncodeScale = 4.0;
//const highp float numberOfDeltasToJump = 6.0;
const int iterativeLimit = 10;


const highp float COLOR_EPSILON = 0.00390625; // 2^-8
const highp float FLOW_EPSILON  = 0.00390625; // 2^-8
const highp float MATRIX_EPSILON  = 0.00390625; // 2^-8


highp vec4 nextColor;



// Common functions

highp float determinant( const highp mat2 matrix )
{
    return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0];
}

highp mat2 inverse( const highp mat2 matrix )
{
    highp float determinant = determinant( matrix );
    
    if( determinant < MATRIX_EPSILON ) return mat2( 0.0 );
    
    highp float oneOverDet = 1.0 / determinant;
    
    return mat2( vec2( matrix[1][1], -matrix[0][1] ), vec2( -matrix[1][0], matrix[0][0] ) ) * oneOverDet;
}



// Encoding

highp vec2 decodeTextureUniformPerFrame( highp vec4 encodedPixelsPerFrame )
{
    highp vec4 textureUniformPerFrame = encodedPixelsPerFrame / textureUniformEncodeScale;
    
    highp vec2 positive = textureUniformPerFrame.xy;
    highp vec2 negative = textureUniformPerFrame.zw;
    
    return positive - negative;
}

highp vec4 encodeTextureUniformPerFrame( highp vec2 textureUniformPerFrame )
{
    highp vec2 scaledTextureUniformPerFrame = textureUniformPerFrame * textureUniformEncodeScale;
    
    highp vec2 positive = clamp(  scaledTextureUniformPerFrame, 0.0, 1.0 );
    highp vec2 negative = clamp( -scaledTextureUniformPerFrame, 0.0, 1.0 );
    
    return vec4( positive, negative );
}

highp vec2 flowAtOffset( highp vec2 offset )
{
    return decodeTextureUniformPerFrame( texture2D(motionTexture,coordinate + offset) );
}

highp vec2 objectiveFunction( highp vec2 currentInverseFlow )
{
    return currentInverseFlow + flowAtOffset( currentInverseFlow );
}



// Check

bool checkInverseFlowColor( highp vec2 inverseFlow )
{
    highp vec4 colorDifference = nextColor - texture2D( framePrev, coordinate + inverseFlow );
    highp float dotProd =  dot(colorDifference,colorDifference);
	if (dotProd < COLOR_EPSILON){
		return true;
	}
	else {
		return false;
	}
		//	return false;
}



// Derivitive

highp vec2 flowDerivitiveXAbout( highp vec2 inverseFlow )
{
    return (objectiveFunction(inverseFlow+vec2( deltaX,      0)) - objectiveFunction(inverseFlow+vec2(-deltaX,      0)) //+
//            objectiveFunction(inverseFlow+vec2( deltaX, deltaY)) - objectiveFunction(inverseFlow+vec2(-deltaX, deltaY))) /
//            (4.0*deltaX);
            ) / (2.0*deltaX);
}

highp vec2 flowDerivitiveYAbout( highp vec2 inverseFlow )
{
    return (objectiveFunction(inverseFlow+vec2(      0, deltaY)) - objectiveFunction(inverseFlow+vec2(      0,-deltaY)) //+
//            objectiveFunction(inverseFlow+vec2( deltaX, deltaY)) - objectiveFunction(inverseFlow+vec2( deltaX,-deltaY))) /
//            (4.0*deltaY);
            ) / (2.0*deltaY);
}



// Search functions and variables

highp vec2 betterLocation;
highp float betterLocationDotProduct;

highp float offsetFlowDotProduct( highp vec2 offset )
{
    return dot( offset, flowAtOffset( offset ) );
}

bool checkForBetter( highp vec2 offsetFromFlow, highp vec2 inverseFlow )
{
    highp vec2 compareOffsetFromCoord = inverseFlow + offsetFromFlow;
    highp float compareOffsetDotProduct = offsetFlowDotProduct( compareOffsetFromCoord );
    
    if( compareOffsetDotProduct < betterLocationDotProduct )
    {
        betterLocation = compareOffsetFromCoord;
        betterLocationDotProduct = compareOffsetDotProduct;
        return true;
    }
    return false;
}

highp vec2 findInverseFlow()
{
    highp vec2 flowAdditiveInverse = -decodeTextureUniformPerFrame( texture2D(motionTexture,coordinate) );
    
    if( checkInverseFlowColor( flowAdditiveInverse ) )
        return flowAdditiveInverse;
    
    highp vec2 inverseFlow = flowAdditiveInverse;
    
    for( int i = 0; i < iterativeLimit; ++i )
    {
        highp vec2 inverseFlowAdjustment;
        {
            highp vec2 flowDx = flowDerivitiveXAbout( inverseFlow );
            highp vec2 flowDy = flowDerivitiveYAbout( inverseFlow );
            
            highp mat2 flowGradiant = mat2( flowDx, flowDy );
            
            highp mat2 flowGradiantInverse = inverse( flowGradiant );
            
            inverseFlowAdjustment = - flowGradiantInverse * objectiveFunction(inverseFlow);
        }
        
        if( dot(inverseFlowAdjustment,inverseFlowAdjustment) < FLOW_EPSILON )
        {
            highp vec4 colorDifferenceAdditiveInverse =
                nextColor - texture2D( framePrev, coordinate + flowAdditiveInverse );
            highp vec4 colorDifferenceInverseFlow =
                nextColor - texture2D( framePrev, coordinate + inverseFlow );
            
            highp float additiveInversError = dot( colorDifferenceAdditiveInverse, colorDifferenceAdditiveInverse );
            highp float inverseFlowError = dot( colorDifferenceInverseFlow, colorDifferenceInverseFlow );
            
            if( inverseFlowError < additiveInversError )
                return inverseFlow;
            
            return flowAdditiveInverse;
        }
        else
            inverseFlow += inverseFlowAdjustment;
        
        if( checkInverseFlowColor( inverseFlow ) )
            return inverseFlow;
    }
    
    // No appropriate inverse flow was found. Just return the additive inverse.
    return flowAdditiveInverse;
}





void main()
{
    nextColor = texture2D( frameNext, coordinate );
    
    gl_FragColor = encodeTextureUniformPerFrame( findInverseFlow() );
}
