varying highp vec2 coordinate;
uniform sampler2D framePrev;
uniform sampler2D frameNext;
uniform sampler2D dxTexture;
uniform sampler2D dyTexture;
uniform sampler2D prevInverseFlow;
uniform highp float prevInverseFlowExistFlag;
uniform highp float width;
uniform highp float height;

const highp float textureUniformEncodeScale = 4.0;
const int neighborhoodHalfSize = 6;
//const int lucasKanadeIterativeLimit = 2;


highp vec2 pixelSize;


const highp float EPSILON = 0.00390625; // 2^-8
const highp float ACCURACY_THRESHOLD = 0.03;



// Common functions

highp float determinant( const highp mat2 matrix )
{
    return matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0];
}

highp mat2 inverse( const highp mat2 matrix )
{
    highp float determinant = determinant( matrix );
    
    if( determinant < EPSILON ) return mat2( 0.0 );
    
    highp float oneOverDet = 1.0 / determinant;
    
    return mat2( vec2( matrix[1][1], -matrix[0][1] ), vec2( -matrix[1][0], matrix[0][0] ) ) * oneOverDet;
}



// Encoding
highp vec2 decodeGuess()
{
    highp vec4 encodedUniformPerFrame = texture2D(prevInverseFlow,coordinate)*prevInverseFlowExistFlag;
    
    highp vec4 textureUniformPerFrame = encodedUniformPerFrame/textureUniformEncodeScale;
    
    highp vec2 positive = textureUniformPerFrame.xy;
    highp vec2 negative = textureUniformPerFrame.zw;
    
    highp vec2 prevInverseFlowDecoded = positive-negative;
    
    highp vec2 boarderSize = pixelSize*float(neighborhoodHalfSize*2+1);
    
    if( (coordinate.x < boarderSize.x) || (coordinate.x > (1.0-boarderSize.x)) )
        prevInverseFlowDecoded[0] = 0.0;
        
    if( (coordinate.y < boarderSize.y) || (coordinate.y > (1.0-boarderSize.y)) )
        prevInverseFlowDecoded[1] = 0.0;
    
    return -prevInverseFlowDecoded;
}

highp vec4 encodeTextureUniformPerFrame( highp vec2 textureUniformPerFrame )
{
    highp vec2 scaledTextureUniformPerFrame = textureUniformPerFrame * textureUniformEncodeScale;
    
    highp vec2 positive = clamp(  scaledTextureUniformPerFrame, 0.0, 1.0 );
    highp vec2 negative = clamp( -scaledTextureUniformPerFrame, 0.0, 1.0 );
    
    return vec4( positive, negative );
}



// Derivitive

const highp vec4 DERIVITIVE_BIAS = vec4( 0.5, 0.5, 0.5, 0.5 );



// Coordinate functions

highp vec2 coordinateAtOffset( int x, int y )
{
    return coordinate + vec2( x, y ) * pixelSize;
}



// Main

void main()
{
    pixelSize = vec2( 1.0 / width, 1.0 / height );
    
    // Calculate the spatial gradient matrix in pixels squared per color squared
    // Also check if the guess is good or not
    
    highp vec2 guessUniform = decodeGuess();
    
    highp mat2 spatialGradientMatrixInverseP2PC2;
    {
        highp mat2 spatialGradientMatrixC2PP2 = mat2( 0.0 );
        
        
        highp float noMoveError = 0.0;
        highp float guessError = 0.0;
        
        for( int x = -neighborhoodHalfSize; x <= neighborhoodHalfSize; ++x )
        {
            for( int y = -neighborhoodHalfSize; y <= neighborhoodHalfSize; ++y )
            {
                highp vec2 atCoordUniform = coordinateAtOffset( x, y );
                
                highp vec4 dxCPP = texture2D(dxTexture,atCoordUniform) - DERIVITIVE_BIAS;
                highp vec4 dyCPP = texture2D(dyTexture,atCoordUniform) - DERIVITIVE_BIAS;
                
                spatialGradientMatrixC2PP2[0][0] += dot(dxCPP,dxCPP);
                spatialGradientMatrixC2PP2[0][1] += dot(dxCPP,dyCPP);
                spatialGradientMatrixC2PP2[1][1] += dot(dyCPP,dyCPP);
                
                
                highp vec4 previousColor = texture2D(framePrev,atCoordUniform);
                highp vec4 noMoveColor   = texture2D(frameNext,atCoordUniform);
                highp vec4 guessColor    = texture2D(frameNext,atCoordUniform+guessUniform);
                
                highp vec4 noMoveDifference = previousColor - noMoveColor;
                highp vec4 guessDifference  = previousColor -  guessColor;
                
                noMoveError += dot(noMoveDifference,noMoveDifference);
                guessError  += dot( guessDifference, guessDifference);
            }
        }
        spatialGradientMatrixC2PP2[1][0] = spatialGradientMatrixC2PP2[0][1];
        
        spatialGradientMatrixInverseP2PC2 = inverse( spatialGradientMatrixC2PP2 );
        
        
        if(noMoveError<guessError)
            guessUniform = vec2(0.0,0.0);
    }
    
//    for( int k = 0; k < lucasKanadeIterativeLimit; ++k )
//    {
        highp vec2 imageMismatchVectorC2PP = vec2(0.0,0.0);
        
        for( int x = -neighborhoodHalfSize; x <= neighborhoodHalfSize; ++x )
        {
            for( int y = -neighborhoodHalfSize; y <= neighborhoodHalfSize; ++y )
            {
                highp vec2 atCoordUniform = coordinateAtOffset( x, y );
                
                highp vec4 imageDifferenceColor = texture2D(framePrev,atCoordUniform) - texture2D(frameNext,atCoordUniform+guessUniform);
                
                highp vec4 dxCPP = texture2D(dxTexture,atCoordUniform) - DERIVITIVE_BIAS;
                highp vec4 dyCPP = texture2D(dyTexture,atCoordUniform) - DERIVITIVE_BIAS;
                
                imageMismatchVectorC2PP[0] += dot(imageDifferenceColor,dxCPP);
                imageMismatchVectorC2PP[1] += dot(imageDifferenceColor,dyCPP);
            }
        }
        
        highp vec2 adjustmentPixels = spatialGradientMatrixInverseP2PC2 * imageMismatchVectorC2PP;
        
//        if( dot(adjustmentPixels,adjustmentPixels) < ACCURACY_THRESHOLD )
//            break;
        
        guessUniform += adjustmentPixels * pixelSize;
//    }
    
    gl_FragColor = encodeTextureUniformPerFrame( guessUniform );
}

